#include "hourglass.h"

double MHZ = 0;
int CLOCK_HZ = 1024;
int num_threads = 0;
int duration;

struct thrd_data_t *thrd_data = NULL;
struct trace_rec_t *trace_recs = NULL;
int current_record;
int tracing;
cycle_time start, end;

static void init_thrd_data (int x)
{
  int i;

  num_threads = x;

  thrd_data = (struct thrd_data_t *) 
    xmalloc (x * sizeof (struct thrd_data_t));

  for (i=0; i<x; i++) {
    thrd_data[i].pri = NORMAL;
    thrd_data[i].lat_slots = NULL;
    thrd_data[i].lat_slot = 0;
    thrd_data[i].exec_start = 0;
    thrd_data[i].last_exec = 0;
    thrd_data[i].reserved_record = -1;
  }
}

static void usage (void)
{
  printf ("usage: lat | trace1\n");
  printf ("  trace1  -- single thread execution trace\n");
  printf ("  trace10 -- ten thread execution trace\n");
  printf ("  lat     -- single thread scheduling latency test\n");
  exit (-1);
}

double gap_us;
cycle_time GAP;
int *gap_hist;

static void parse_args (int argc, char *argv[])
{
  int i;

  if (argc != 2) {
    usage ();
  }

  gap_us = 2.0;
  
  if (strcmp (argv[1], "lat") == 0) {

    duration = 10000000;
    init_thrd_data (1);
    thrd_data[0].task = LAT;
    thrd_data[0].pri = HIGHEST;
    thrd_data[0].lat_slots = (cycle_time *)
      xmalloc (sizeof (cycle_time) * LAT_SLOTS);

  } else if (strcmp (argv[1], "trace1") == 0) {    

    duration = 10000000;
    init_thrd_data (1);
    thrd_data[0].task = TRACE;

  } else if (strcmp (argv[1], "trace10") == 0) {    

    duration = 10000000;
    init_thrd_data (10);
    for (i=0; i<10; i++) {
      thrd_data[i].task = TRACE;
    }

  } else {
    usage ();
  }

  tracing = 0;
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].task == TRACE) {
      tracing = 1;
    }
  }

  if (tracing) {
    trace_recs = (struct trace_rec_t *)
      xmalloc (sizeof (struct trace_rec_t) * NUM_TRACE_RECORDS);
    gap_hist = (int *) 
      xmalloc (sizeof (int) * NUM_GAP_HIST_BUCKETS);
    
    for (i=0; i<NUM_GAP_HIST_BUCKETS; i++) {
      gap_hist[i] = 0;
    }

  } else {
    trace_recs = NULL;
    gap_hist = NULL;
  }
  current_record = 0;
}

static void wait_for_all_threads (thrd_status_t new_status)
{
  int cnt = 0;
  int ready;
  int i;

  do {
    if (cnt++ > 1000) {
      printf ("Hrm: waited too long for threads to reach state %d\n",
	      new_status);
      exit (-1);
    }
    /* argh - usleep of <1000 will busy wait on 2.4.12 */
    usleep (5000);
    ready = 1;
    for (i=0; i<num_threads; i++) {
      if (thrd_data[i].status != new_status) {
	ready = 0;
      }
    }
  } while (!ready);
}

static void set_all_threads (thrd_status_t new_status)
{
  int i;

  for (i=0; i<num_threads; i++) {
    thrd_data[i].status = new_status;
  }
}

static void flush_records (void)
{
  int i, rec;

  for (i=0; i<num_threads; i++) {
    rec = thrd_data[i].reserved_record;
    if (rec < 0) {
      continue;
    }
    trace_recs[rec].id = i;
    trace_recs[rec].start = thrd_data[i].exec_start;
    trace_recs[rec].end = thrd_data[i].last_exec;
  }
}
  
int main (int argc, char *argv[])
{
  int i, retval;

  // time conversion functions won't work until this runs
  MHZ = find_mhz ();

  parse_args (argc, argv);

  if (tracing) {
    GAP = us_to_cycle (gap_us);
    printf ("max gap is %lld cycles\n", GAP);
  }

  if (num_threads == 0) {
    printf ("Hrm: internal error\n");
    exit (-1);
  }

  lock_pages ();
  set_priority (MAX);
  setup_fast_clock ();

  set_all_threads (INIT);

  for (i=0; i<num_threads; i++) {
    retval = pthread_create (&thrd_data[i].thrd,
			     NULL,
			     thrd_func,
			     (void *)i);
    if (retval != 0) {
      printf ("Hrm: error creating thread %d\n", i);
      exit (-1);
    }
  }

  wait_for_all_threads (READY);
  set_all_threads (GO);
  start = rdtsc();

  usleep (duration);

  set_all_threads (STOP);
  wait_for_all_threads (DONE);
  end = rdtsc();
  if (tracing) {
    flush_records ();
  }
  set_priority (NORMAL);
  unset_fast_clock ();
  print_all_results ();

  return 0;
} 
