#include "hourglass.h"

static void lat_test (int id)
{
  cycle_time now, expire;
  unsigned long stuff;
  int retval;

  while (thrd_data[id].status != STOP) {
    expire = rdtsc() + sec_to_cycle ((double)1 / CLOCK_HZ);
    retval = read (RTCLOCKFD, &stuff, sizeof (unsigned long));
    if (retval == -1) {
      perror("read");
      exit(errno);
    }
    now = rdtsc();
    
    if (thrd_data[id].lat_slot < LAT_SLOTS) {
      thrd_data[id].lat_slots[thrd_data[id].lat_slot] = now - expire;
      thrd_data[id].lat_slot++;
    }
  }
}

static int get_new_rec (void)
{
  int n = current_record;

  // FIXME! needs to be atomic
  current_record++;

  if (current_record >= NUM_TRACE_RECORDS) {
    printf ("oops: trace record overrun - exiting\n");
    exit (-1);
  }
  return n;
}

static void poll (int id)
{
  cycle_time t;
  cycle_time gap;

  t = rdtsc ();
  gap = t - thrd_data[id].last_exec;

  if (gap > GAP) {
    if (thrd_data[id].exec_start == 0) {
      thrd_data[id].reserved_record = get_new_rec ();
    } else {
      int rec = thrd_data[id].reserved_record;
      thrd_data[id].reserved_record = get_new_rec ();
      trace_recs[rec].id = id;
      trace_recs[rec].start = thrd_data[id].exec_start;
      trace_recs[rec].end = thrd_data[id].last_exec;
    }
    thrd_data[id].exec_start = t;
  } else {
    if (gap/GAP_HIST_DIVISOR > NUM_GAP_HIST_BUCKETS) {
      printf ("hrm: gap %lld exceeds NUM_GAP_HIST_BUCKETS %d\n",
	      gap/GAP_HIST_DIVISOR, NUM_GAP_HIST_BUCKETS);
      exit (-1);
    }
    gap_hist[gap/GAP_HIST_DIVISOR]++;
  }

  thrd_data[id].last_exec = t;
}

static void trace (int id)
{
  while (thrd_data[id].status != STOP) {
    poll (id);
  }
}

void *thrd_func (void *arg)
{
  int id = (int)arg;

  set_priority (thrd_data[id].pri);

  thrd_data[id].status = READY;

  do { } while (thrd_data[id].status != GO);

  switch (thrd_data[id].task) {
    
  case LAT:
    lat_test (id);
    break;
    
  case TRACE:
    trace (id);
    break;
    
  default:
    printf ("Hrm: thread %d has an unknown task\n", id);
    exit (-1);
  }

  thrd_data[id].status = DONE;

  return NULL;
}
